/*
 * Created on 11.07.2003
 *
 */

package com.lauterbach.trace32.api;

/**
 * @author Rudi Dienstbeck
 * @author Oliver Oppitz
 */

import general.utils.Utils;

import java.util.*;
import java.io.File;

public class Trace32API {

	HashMap knownBreaks = new HashMap(); /* Local list of known breaks; is updated now and then to reflect actual breaks in T32 */ 

	Vector breakpointListeners = new Vector();

	static final String echoString = "ECLIPSE: "; /* added to echo of each
												   * PRACTICE command; can be
												   * adapted for individual
												   * needs.
												   */

	public static final int DEVICE_OS = 0;

	public static final int DEVICE_ICE = 1;

	public static final int STATE_DOWN = 0;

	public static final int STATE_SLEEP = 1;

	public static final int STATE_STOPPED = 2;

	public static final int STATE_RUNNING = 3;

	public static final int STEPMODE_INTO = 0;

	public static final int STEPMODE_OVER = 0x80;

	public static final int STEPMODE_ASM = 0;

	public static final int STEPMODE_HLL = 0x01;

	private Trace32Line line;

	private byte msgSeqNum = 0;

	private boolean keepPolling = false;

	private Thread pollThread = null;

	int pollDelayMs; // Delay for polling thread

	static Trace32API interfaceObjects[] = new Trace32API[100];

	static int interfaceObjectCount = 0;

	/**
	 * Constructor is private because instances are only created through
	 * getTraceAPI() calls
	 */
	private Trace32API(int port) {
		line = new Trace32Line(port);
	}
	
	/**
	 * Creates a Trace32API object and connects it to the debugger. The returned
	 * object is "ready-for-use".
	 * 
	 * @return Will always return the same object for a given value of port.
	 */
	
	public static Trace32API getTrace32API(int port) throws Trace32Exception {

		for (int i = 0; i < interfaceObjectCount; i++) {
			if (interfaceObjects[i].line.getTxPort() == port) {
				return interfaceObjects[i];
			}
		}
		Trace32API newAPI = new Trace32API(port);
//		interfaceObjects[interfaceObjectCount++] = newAPI;

		newAPI.init();
		

		try {
			newAPI.attach();
		} catch (Trace32Exception e) {

			/*
			 * If attach fails, try a second time. This happens in case the
			 * previous program did not call exit() as is frequent during
			 * debugging the Java code.
			 */
//			System.out
//					.println("DEBUG: AAttaching to Trace32.JAPI failed --> retrying");
			newAPI.exit();
			newAPI.init();
			newAPI.attach(); /* if second attempt fails, throws an exception */
//			System.out.println("DEBUG: Attached to Trace32.JAPI successfully.");
		}
		interfaceObjects[interfaceObjectCount++] = newAPI;
		return newAPI;
	}

	/**
	 * Terminates connection with debugger. Call before releaseing Trace32API
	 * object.
	 */
	public synchronized void exit() throws Trace32Exception {

		enablePolling(false, 0); /* disable polling thread, just in case */
		line.exit();
	}

	/**
	 * * Configures the interface object.
	 * 
	 * @deprecated use getTraceAPI() instead
	 * 
	 * @param key
	 *            "PORT=", "HOSTPORT=", "PACKLEN=", "TIMEOUT="
	 * @param value
	 */

	void configure(String key, String value) {
		line.configure(key, value);
	}

	synchronized void init() throws Trace32Exception {
		line.init();
		line.sync();
	}

	void attach() throws Trace32Exception {
		byte[] buffer = new byte[4];

		buffer[0] = 2;
		buffer[1] = 0x71;
		buffer[2] = DEVICE_ICE;
		buffer[3] = ++msgSeqNum;

		line.transmit(buffer, 4);
		line.receive(msgSeqNum);
	}

	/** Check connection with debugger */
	public synchronized void ping() throws Trace32Exception {
		byte[] buffer = new byte[4];

		buffer[0] = 2;
		buffer[1] = 0x73; /* PING */
		buffer[2] = 0;
		buffer[3] = ++msgSeqNum;

		line.transmit(buffer, 4);
		line.receive(msgSeqNum);
	}

	/** Sends a PRACTICE command to T32 for execution. */
	public synchronized void command(String c) throws Trace32Exception {
		if (c.length() >= 256)
			throw new Trace32Exception(Trace32Exception.PARAMETER,
					"command length");

		int cLength = c.length();
		int bLength = 4 + cLength + 1; // header + command + 0
		if ((bLength & 1) == 1) // ensure word alignment
			bLength++;

		byte[] buffer = new byte[bLength];

		buffer[0] = (byte) (cLength + 3);
		buffer[1] = 0x72; /* generic practice command */
		buffer[2] = 0x02;
		buffer[3] = ++msgSeqNum;
		System.arraycopy(c.getBytes(), 0, buffer, 4, cLength);
		buffer[4 + cLength] = 0;

		line.transmit(buffer, bLength);
		line.receive(msgSeqNum);
	}

	/** Starts the target application. */
	public synchronized void go() throws Trace32Exception {
		byte[] buffer = new byte[4];

		buffer[0] = 2;
		buffer[1] = 0x74;
		buffer[2] = 0x51;
		buffer[3] = ++msgSeqNum;

		line.transmit(buffer, 4);
		line.receive(msgSeqNum);
	}

	/** Stops the target application. */
	public synchronized void stop() throws Trace32Exception {
		byte[] buffer = new byte[4];

		buffer[0] = 2;
		buffer[1] = 0x74;
		buffer[2] = 0x52;
		buffer[3] = ++msgSeqNum;

		line.transmit(buffer, 4);
		line.receive(msgSeqNum);
	}

	/**
	 * Single steps target application.
	 * 
	 * @param mode:
	 *            0=assembler step, 1=HLL step, 2=mixed mode
	 */
	synchronized public void stepMode(int mode) throws Trace32Exception {
		byte[] buffer = new byte[6];

		buffer[0] = 2;
		buffer[1] = 0x74;
		buffer[2] = 0x54;
		buffer[3] = ++msgSeqNum;
		buffer[4] = (byte) mode;
		buffer[5] = 0;

		line.transmit(buffer, 6);
		line.receive(msgSeqNum);
	}

	/**
	 * Returns the address associated with a symbol.
	 * 
	 * @param s name of the symbol
	 * @return address of the symbol
	 * @throws Trace32Exception
	 * @TODO change int parameter to something better (address type, long, ...)
	 */
	public synchronized int getSymbolAddress(String s) throws Trace32Exception {
		if (s.length() >= 256)
			throw new Trace32Exception(Trace32Exception.PARAMETER,
					"symbol length");

		int sLength = s.length();
		int bLength = 4 + sLength + 1; // header + symbol + 0
		if ((bLength & 1) == 1) // ensure word alignment
			bLength++;

		byte[] buffer = new byte[bLength];

		buffer[0] = (byte) (sLength + 3);
		buffer[1] = 0x74;
		buffer[2] = 0x62;
		buffer[3] = ++msgSeqNum;
		System.arraycopy(s.getBytes(), 0, buffer, 4, sLength);
		buffer[4 + sLength] = 0;

		line.transmit(buffer, bLength);
		buffer = line.receive(msgSeqNum);

		return Trace32Line.getLongFromBuffer(buffer, 5);
	}

	/**
	 * Reads the last line of the area window; used for detecting errors when
	 * executing special practice commands.
	 */

	public synchronized String getMessage() throws Trace32Exception {
		byte[] buffer = new byte[4];

		buffer[0] = 2;
		buffer[1] = 0x76;
		buffer[2] = 0;
		buffer[3] = ++msgSeqNum;

		line.transmit(buffer, 4);
		buffer = line.receive(msgSeqNum);

		return new String(buffer, 9, buffer.length - 10);
	}

	public synchronized int getMessageMode() throws Trace32Exception {
		byte[] buffer = new byte[4];

		buffer[0] = 2;
		buffer[1] = 0x76;
		buffer[2] = 0;
		buffer[3] = ++msgSeqNum;

		line.transmit(buffer, 4);
		buffer = line.receive(msgSeqNum);

		return Trace32Line.getLongFromBuffer(buffer, 5);
	}

	/**
	 * Returns main state of debugger.
	 * 
	 * @return 0: system is down, not prepared for download or emulation 
	 *         1: system is halted, CPU makes no cycles 
	 *         2: target program is stopped 
	 *         3: target program is running
	 * 
	 * @throws Trace32Exception
	 */
	public synchronized int getState() throws Trace32Exception {
		byte[] buffer = new byte[4];

		buffer[0] = 2;
		buffer[1] = 0x74;
		buffer[2] = 0x10;
		buffer[3] = ++msgSeqNum;

		line.transmit(buffer, 4);
		buffer = line.receive(msgSeqNum);

		return buffer[5];
	}

	/**
	 * Checks whether a PRACTICE script is currently running. Use this method
	 * when waiting for a script to terminate. Do not issue commands until the
	 * script is finished.
	 */
	public synchronized boolean isPracticeRunning() throws Trace32Exception {
		byte[] buffer = new byte[4];

		buffer[0] = 2;
		buffer[1] = 0x72;
		buffer[2] = 0x03;
		buffer[3] = ++msgSeqNum;

		line.transmit(buffer, 4);
		buffer = line.receive(msgSeqNum);

		return (buffer[5] == 1);
	}
	
 /* *********Inizio delle aggiunte di  merlin********** */

	
	// steppa passo passo (provata: FUNZIONA!!!)
	/** Steppa passo passo 
	 * @throws Trace32Exception
	 */
	public void Step() throws Trace32Exception {
		byte[] b= new byte[128];
		byte[] b2 = new byte[128]; //buffer di ritorno
		b[0]=2;
		b[1]=0x74;
		b[2]=(byte) 0x50;
		b[3]=++msgSeqNum;
		
		line.transmit(b,4);
		b2=line.receive(msgSeqNum);
		
	}
	
	
	/** Restituisce lo stato dell'Analyzer
	 * @param state
	 * @param size
	 * @param min
	 * @param max
	 * @return
	 * @throws Trace32Exception
	 */
	public int AnaStatusGet(byte state, long size, long min, long max) throws Trace32Exception {
		
		byte[] b= new byte[128];
		byte[] b2 = new byte[128]; //buffer di ritorno
		int i;
		String s;
		
		b[0]=2;
		b[1]=0x74;
		b[2]=(byte) 128;
		b[3]=++msgSeqNum;
			
		line.transmit(b,4);
		b2=line.receive(msgSeqNum);
		
		state=b2[8];
		
		// estraggo la dimensione
		//s="";
		//for(i=12;i>=9;i--)
		//	s +=Utils.toHex(b[i]);
		//size=Long.parseLong(s,16);
		
		size= Utils.toIntegerLE(b2,9);
		
		// estraggo il minimo
		//s="";
		//for(i=16;i>=13;i--)
		//	s +=Utils.toHex(b[i]);
		//min=Long.parseLong(s,16);
		
		min= Utils.toIntegerLE(b2,13);
		// estraggo il massimo
		//s="";
		//for(i=20;i>=17;i--)
		//	s +=Utils.toHex(b[i]);
		//max=Long.parseLong(s,16);
		
		max= Utils.toIntegerLE(b2,17);
		// ritorno 0 se tutto OK!
		return 0;
//		System.out.println("-----Buffer di ritorno della AnaStatusGet!-------");
//		System.out.println("State: "+toHex(b2[8]));
//		System.out.println("Size: "+toHex(b2[12])+toHex(b2[11])+toHex(b2[10])+toHex(b2[9]));
//		System.out.println("Min: "+toHex(b2[16])+toHex(b2[15])+toHex(b2[14])+toHex(b2[13]));
//		System.out.println("Max: "+toHex(b2[20])+toHex(b2[19])+toHex(b2[18])+toHex(b2[17]));
//		System.out.println("------Fine buffer di ritorno di AnaStatusget!------");
		// byte 8 		: state
		// byte 9-12	: size
		// byte 13-16 	: min
		// byte 17-20 	: max
		
		
	}
	
/** Ritorna il max valore dell'indice dei record nell'analyzer
 * @return
 * @throws Trace32Exception
 */
public long AnaStatusMax() throws Trace32Exception {
		
		byte[] b= new byte[128];
		byte[] b2 = new byte[128]; //buffer di ritorno
				
		b[0]=2;
		b[1]=0x74;
		b[2]=(byte) 0x80;
		b[3]=++msgSeqNum;
			
		line.transmit(b,4);
		b2=line.receive(msgSeqNum);
		int max=0;
		
		//byte[] max_array=new byte[4];
		//System.arraycopy(b2,17,max_array,0,4);
		max= Utils.toIntegerLE(b2,17);
		
		return max;
	}

/** Ritorna il min valore dell'indice dei record nell'analyzer
 * @return
 * @throws Trace32Exception
 */
public long AnaStatusMin() throws Trace32Exception {
		
		byte[] b= new byte[128];
		byte[] b2 = new byte[128]; //buffer di ritorno
				
		b[0]=2;
		b[1]=0x74;
		b[2]=(byte) 128;
		b[3]=++msgSeqNum;
			
		line.transmit(b,4);
		b2=line.receive(msgSeqNum);
		int min=0;
		// estraggo il minimo
		//byte[] min_array=new byte[4];
		//System.arraycopy(b2,13,min_array,0,4);
		min= Utils.toIntegerLE(b2,13);
		

		return min;
	}


/** Ritorna la dimensione del buffer dell'analyzer
 * @return
 * @throws Trace32Exception
 */
public long AnaStatusSize() throws Trace32Exception {
		
		byte[] b= new byte[128];
		byte[] b2 = new byte[128]; //buffer di ritorno
				
		b[0]=2;
		b[1]=0x74;
		b[2]=(byte) 128;
		b[3]=++msgSeqNum;
			
		line.transmit(b,4);
		b2=line.receive(msgSeqNum);
		int size=0;

		//byte[] size_array=new byte[4];
		//System.arraycopy(b2,9,size_array,0,4);
		size= Utils.toIntegerLE(b2,9);
		
		return size;
	}
	
	/** Restituisce un buffer che contiente le info delrecord specificato
	 * @param nrecord				num del record da predenre
	 * @param buff					buffer di ritorno su cui copiare le info
	 * @param len					lunghezzza del buffer
	 * @throws Trace32Exception		riferimento al tracer
	 */
	public void AnaRecordGet(int nrecord, byte[]  buff, int len) throws Trace32Exception {
		byte[] buffer = new byte[len+5];
		byte[] buf2=new byte[len]; //buffer di ritorno
		String snum;
		String zeros="";
		
		
		buffer[0]= (byte)8;
		buffer[1]= (byte) 0x74;
		buffer[2]= (byte) 0x81;
		buffer[3]=(byte) ++msgSeqNum;
		
		/*System.err.println("Sono all'interno di AnaRecordGet!");
		System.err.println("Il num di record passato : "+nrecord);
		System.err.println("Ma il vettore :");*/
		
		snum=Integer.toHexString(nrecord);
		for(int k=1;k<=(8-snum.length());k++) 
			zeros += "0";
		snum = zeros+snum;
		
		// metto nei bytes 4-7 il num del record
		
		for(int i=0;i<=7;i+=2) {
			buffer[7-(i/2)]= (byte)Integer.parseInt(snum.substring(i,i+2),16);
			/*System.err.print((7-(i/2)+"\t"+buffer[7-(i/2)]));
			System.err.println("\t"+Integer.toHexString(buffer[7-(i/2)]));
			*/
		}
		//buffer[4]=(byte) nrecord;
		//buffer[5]=(byte) 255;
		//buffer[6]=(byte) 255;
		//buffer[7]=(byte) 255;  // metto nei bytes 4-7 il num del record
		
		buffer[8]=(byte) len;
		buffer[9]=(byte) 0; // metto nei bytes 8-9 la lunghezza
	
		//System.out.println("------AnaRecordGet ("+nrecord+")");
		//System.out.println("Questo  il buffer trasmesso.");
		//for(int k=0;k<10;k++) {
		//	System.out.println(k+"   0x"+ (traduci(buffer[k])));
		//
		//} 
		
		line.transmit(buffer,11); // trasmetto 10 bytes (i bytes da 0 a 9)
		
		buf2=line.receive(msgSeqNum);
		System.arraycopy(buf2,5,buff,0,len);
		
		//System.out.println("-------Inizio del buffer di riitorno----");
		//System.out.println("Ritorno: (dec) "+buf2[0]);
		//System.out.println("Address: 0x"+traduci(buff[7])+traduci(buff[6])+traduci(buff[5])+traduci(buff[4]));
		//for(int j=0;j<60;j++)
		//	System.out.println(j+"    0x"+traduci(buf2[j]));
		//System.out.println("-------Fine del buffer-------");
		//System.out.println("------Fine AnaRecordGet---------------");
		

		//if (buf2[2]!=0)
		//	System.out.println("ERRORE n. "+buf2[2]+" nella AnaRecordGet!");
	
	
	}
	
	/** Restituisce la lunghezza di un simbolo
	 * @param s
	 * @return
	 * @throws Trace32Exception
	 */
	public synchronized int getSymbolLen(String s) throws Trace32Exception {
		if (s.length() >= 256)
			throw new Trace32Exception(Trace32Exception.PARAMETER,
					"symbol length");

		int sLength = s.length();
		int bLength = 4 + sLength + 1; // header + symbol + 0
		if ((bLength & 1) == 1) // ensure word alignment
			bLength++;

		byte[] buffer = new byte[bLength];

		buffer[0] = (byte) (sLength + 3);
		buffer[1] = 0x74;
		buffer[2] = 0x62;
		buffer[3] = ++msgSeqNum;
		System.arraycopy(s.getBytes(), 0, buffer, 4, sLength);
		buffer[4 + sLength] = 0;

		line.transmit(buffer, bLength);
		buffer = line.receive(msgSeqNum);

		return Trace32Line.getLongFromBuffer(buffer, 9);
	}


	
/* ************ Fine delle aggiunte di merlin ************** */	
	
	public synchronized int readPC() throws Trace32Exception {
		byte[] buffer = new byte[4];

		buffer[0] = 2;
		buffer[1] = 0x74;
		buffer[2] = 0x22;
		buffer[3] = ++msgSeqNum;

		line.transmit(buffer, 4);
		buffer = line.receive(msgSeqNum);

		return Trace32Line.getLongFromBuffer(buffer, 5);
	}

	/**
	 * Reads registers from target.
	 * 
	 * @param mask
	 *            defines registers to read; mask is specific for the target
	 *            architecture
	 * @return
	 * @throws Trace32Exception
	 * 
	 * @limitation: mask too small for current architectures
	 */
	public synchronized int[] readRegisters(long mask) throws Trace32Exception {
		byte[] buffer = new byte[12];

		buffer[0] = 12;
		buffer[1] = 0x74;
		buffer[2] = 0x20;
		buffer[3] = ++msgSeqNum;

		Trace32Line.setBufferFromLong(buffer, 4, (int) (mask & 0xffffffff));
		Trace32Line.setBufferFromLong(buffer, 8, (int) (mask >>> 32));

		line.transmit(buffer, 12);
		buffer = line.receive(msgSeqNum);

		int[] regs = new int[64];
		int regIndex = 0;
		int bufIndex = 5;

		while (mask != 0) {
			if ((mask & 1) == 1) {
				regs[regIndex++] = Trace32Line.getLongFromBuffer(buffer, bufIndex);
				bufIndex += 4;
			}
			mask >>>= 1;
		}

		return regs;
	}

	/**
	 * Returns source file corresponding to P:-address.
	 * 
	 * @param address
	 * @return
	 * @throws Trace32Exception
	 * @note uses the same C-API call as getSourceLINE
	 * 
	 * @limitation int-type not suitable for address (32bits)
	 */
	public synchronized String getSourceFile(long address) throws Trace32Exception {
		byte[] buffer = new byte[8];

		buffer[0] = 6;
		buffer[1] = 0x74;
		buffer[2] = 0x60;
		buffer[3] = ++msgSeqNum;
		Trace32Line.setBufferFromLong(buffer, 4, (int)address);

		line.transmit(buffer, 8);
		buffer = line.receive(msgSeqNum);

		return new String(buffer, 9, buffer.length - 10);
	}

	/**
	 * Returns source line corresponding to P:-address.
	 * 
	 * @param address
	 * @return
	 * @throws Trace32Exception
	 * @note uses the same C-API call as getSourceFILE
	 * 
	 * @limitation long-type not suitable for address (32bits)
	 */
	public synchronized int getSourceLine(long address) throws Trace32Exception {
		byte[] buffer = new byte[8];

		buffer[0] = 6;
		buffer[1] = 0x74;
		buffer[2] = 0x60;
		buffer[3] = ++msgSeqNum;
		Trace32Line.setBufferFromLong(buffer, 4, (int) address);

		line.transmit(buffer, 8);
		buffer = line.receive(msgSeqNum);

		return Trace32Line.getLongFromBuffer(buffer, 5);
	}

	
	
	/** ********* Eclipse API extensions ********** */

	/** Adds listener for state modifications */
	public synchronized void addListener(ITrace32Listener listener) {
		breakpointListeners.add(listener);
	}

	/** Removes listener for state modifications */
	public synchronized void removeListener(ITrace32Listener listener) {
		breakpointListeners.remove(listener);
	}

	
	/**
	 * Updates status and informs clients via listener interface.
	 * 
	 * Currently only breakpoints are supported but in future also runstate,
	 * headmode etc. can be included.
	 * 
	 * Compares breakpoint status between T32 and local buffers in Trace32API.
	 * 
	 * This method is the only place where Trace32Breakpoint instances are
	 * created. As breakpoints are complicated objects (in T32) this can only be
	 * done reliably and consistently when T32 reports the exact breakpoint data
	 * and we build the object based on this report.
	 * 
	 * The implementation currently reads complete breakpoint status from
	 * debugger. This may change and be replaced by a low-level notification
	 * mechanism.
	 */

	boolean activeUpdateStatus = false;
	boolean reenteredUpdateStatus = false;

	
	public synchronized void updateStatus() throws Trace32Exception {
		
		/* If reentered set flag and exit. */
		if (activeUpdateStatus) {
			reenteredUpdateStatus = true;
			return;		
		}
		
		activeUpdateStatus = true; 
		
		/* Get changes from T32 and send notifications to listeners. */
		updateStatusHelper(); 	
					
		/*
		 * If - in the meantime - notifications (or other events) created
		 * recursive calls to updateStatus(), call helper again to report these
		 * (potential) changes.
		 */
		if (reenteredUpdateStatus) {
			updateStatusHelper(); 	
		}
		
		activeUpdateStatus = reenteredUpdateStatus = false;
		
	}


	synchronized void updateStatusHelper() throws Trace32Exception {
		
		HashMap newMap = getRemoteBreakpointList();

		/* Find deleted breakpoints */
		Iterator iterKnown = knownBreaks.values().iterator();
		while (iterKnown.hasNext()) {
			Trace32Breakpoint knownBreak = (Trace32Breakpoint)iterKnown.next();
			Long address = new Long(knownBreak.getAddress());
			if (!newMap.containsKey(address)) {
				reportDeleted(knownBreak);
				iterKnown.remove(); /* remove the break from the MAP; not valid to call HashMap.remove! */  
			}
		}

		/* Check for new and modified breakpoints */
		Iterator iterNewMap = newMap.values().iterator();
		while (iterNewMap.hasNext()) {
			Trace32Breakpoint newBreak = (Trace32Breakpoint) (iterNewMap.next());
			Long address = new Long(newBreak.getAddress());
			if (knownBreaks.containsKey(address)) { /* break on same adress? */
				Trace32Breakpoint oldBreak = (Trace32Breakpoint) knownBreaks.get(address);
				if (!oldBreak.equals(newBreak)) { /* modified breakpoint */
					knownBreaks.put(address, newBreak);
					reportDeleted(oldBreak);
					reportCreated(newBreak);
				}
			} else { //new breakpoint
				knownBreaks.put(address, newBreak);
				reportCreated(newBreak); //not found list -> report created
			}
		}
	}
	
	/**
	 * Enables or disables polling for status updates from T32.
	 * 
	 * As the polling is done in a separate thread it is asynchronous to the
	 * main application. If this is undesireable, call reportBreakStatus() from
	 * the main thread when appropriate.
	 * 
	 * @param enable
	 *            state of the polling thread.
	 * @param msdelay
	 *            delay in ms to elapse before next polling. Typically uses
	 *            1000ms.
	 * @throws Trace32Exception
	 * 
	 * @note Currently the polling is required to get updates from T32. It may
	 *       become superfluous in future versions of the Trace32API due to a
	 *       low-level notification mechanism. In that case, enablePolling()
	 *       will become an empty method.
	 */

	public synchronized void enablePolling(boolean enable, int msdelay)
			throws Trace32Exception {

		keepPolling = enable;
		pollDelayMs = msdelay;

		if (enable && (pollThread == null)) {

			pollThread = new Thread() {
				public void run() {
					while (keepPolling) {
						try {
							updateStatus();
							//	dumpAllBreakpoints(); /* DEBUG only */
							sleep(pollDelayMs);
						} catch (InterruptedException e) {
						} catch (Trace32Exception e) {
						}
					}
				}
			};
			pollThread.start();
		}

		if (keepPolling == false) {
			pollThread = null;
		}
	}

	
	/**
	 * Breakpoint handling in T32 is quite complex (e.g. automatic merging of
	 * multiple breaks on same HLL line, moving break to valid line, ...).
	 * Therefore the information about new/changed/deleted breakpoints often
	 * cannot be calculated by Trace32API reliably. In that case the API
	 * reads the new state from T32 via reportBreakStatus() and creates 
	 * notifications for changes. Use addBreakpointListener() to register
	 * a listener.
	 * 
	 * Changes regarding a breakpoint may be reported via reportModified() or in
	 * special cases via reportDeleted() + reportCreated().
	 * 
	 * When changes occur due to a method call in Trace32API (e.g.
	 * enableBreakpoint), the notifications are sent synchronously with the
	 * method.
	 * 
	 * An application may poll for changes through reportStatus() when it
	 * is convenient. Alternatively, enablePeriodicPolling() can be used to get
	 * automatic notifications about changes.
	 * 
	 * The current implementation of state updates requires polling at the lower
	 * levels. This might change in future implementations. This will be
	 * transparent to users of Trace32API.
	 * 
	 * Note that not all breakpoints are delivered to Eclipse e.g. temporary 
	 * breakpoints and stepping breakpoints (used for single stepping instructions 
	 * or HLL lines) are not reported. The general idea is to only report breakpoints
	 * that Eclipse can also create. The reason for this is, that Eclipse will call 
	 * "break.set" even for breaks that set in T32 and reported to Eclipse. This is
	 * due to the Eclipse BreakpointManager implementation. (ecr010, ecr013)  
	 */

	/**
	 * Sets breakpoint in the target application. Calls listener interface to
	 * report changes (e.g. new breakpoint).
	 * 
	 * 
	 * @throws Trace32Exception
	 *             if break cannot be set (wrong file name, line number not
	 *             valid, too many breaks, ...
	 * 
	 * @limitation maps file name to MODULE name by removing suffix (e.g. .cpp,
	 *             .h, .s); will work in vast majority of cases but there is no
	 *             guarantee.
	 */

	public synchronized void setBreakpoint(String pathAndFileName, int lineNr)
			throws Trace32Exception {

		String moduleName = getModuleName(pathAndFileName);
		String moduleLinenr = "\\" + moduleName + "\\"
				+ Integer.toString(lineNr);
		String pcommand = "break.set " + moduleLinenr;

		try {
			checkedCommand(pcommand); /* set the break in debugger */

		} catch (Trace32Exception e) {

			/*
			 * When setting break fails, we need to delete the break again;
			 * deleting non-existing break does no harm
			 */
			System.out.println("DEBUG: failed: " + pcommand);

			pcommand = "break.delete " + moduleLinenr;
			uncheckedCommand(pcommand, false);
			return;
		}

		updateStatus(); // Get the updated state from T32; triggers notifications.
						 
	}

	/**
	 * Tells T32 to delete breakpoint; creates notification regarding deletion.
	 * 
	 * @param bp breakpoint to delete
	 * @throws Trace32Exception
	 */
	public synchronized void deleteBreakpoint(Trace32Breakpoint bp)
			throws Trace32Exception {

		String moduleName = getModuleName(bp.getFileName());
		String moduleLinenr = "\\" + moduleName + "\\"
				+ Integer.toString(bp.getLineNumber());
		String pcommand = "b.delete " + moduleLinenr;
		checkedCommand(pcommand);

		updateStatus(); // Get the updated state from T32; triggers notifications.
	}

	/**
	 * Tells T32 to enable/disable a breakpoint; creates notification regarding
	 * change.
	 * 
	 * @param bp
	 *            breakpoint to disable/enable
	 * @param enable
	 *            the requested state
	 * @throws Trace32Exception
	 */
	public synchronized void enableBreakpoint(Trace32Breakpoint bp,	boolean enable) throws Trace32Exception {

		String moduleName = getModuleName(bp.getFileName());
		String moduleLinenr = "\\" + moduleName + "\\"
				+ Integer.toString(bp.getLineNumber());
		String pcommand = (enable ? "b.enable " : "b.disable ") + moduleLinenr;
		checkedCommand(pcommand);

		updateStatus(); // Get the updated state from T32; triggers notifications.
	}

	
	/** Returns iterator for currently cached breakpoints.
	 * 
	 * @return currently cached breakpoints.
	 */
	
	public synchronized Trace32Breakpoint[] getBreakpoints() throws Trace32Exception {
	
		int i = 0;
		HashMap m = (HashMap)knownBreaks.clone();  /* getRemoteBreakpointList(); */
				
		Trace32Breakpoint[] b = new Trace32Breakpoint[m.size()] ;
		Iterator iter = m.values().iterator();
		while (iter.hasNext()) {
			b[i++]= (Trace32Breakpoint)(iter.next());
		}
		
		return b;
	}
	
	/**
	 * Reads all currently set breakpoints from T32. The delta between this list
	 * and the previously known breakpoints is reported via notifications in
	 * updateStatus().
	 * 
	 * @return all breakpoints set in T32
	 */

	synchronized HashMap getRemoteBreakpointList() throws Trace32Exception {

		HashMap remoteBreaks = new HashMap();  
		
		byte[] txBuffer = new byte[Trace32Line.PKG_LENMAX];
		byte[] rxBuffer;
 	
		txBuffer[0] = 5;            /* size of packet */
		txBuffer[1] = 0x74;         /* Device-Specific Cmd */
		txBuffer[2] = 0x64;         /* function code */
		txBuffer[3] = ++msgSeqNum;  /* message number? info specific for each function? */
		txBuffer[4] = 0x77;         /* sample payload */
		txBuffer[5] = 0;            /* even byte count */
		 	
		line.transmit(txBuffer, 6); /* needs to be even number */
		rxBuffer = line.receive(msgSeqNum);
		 	 
		int rcount = 5; /* skip header */
		 	
		int breakcount = Trace32Line.getWordFromBuffer(rxBuffer, rcount);
		rcount+=2;
		 	
		for(int i=0; i<breakcount; i++){
			long byte_address = Trace32Line.getLongFromBuffer(rxBuffer,rcount);
			rcount+=4;
				
			boolean enabled = ( (rxBuffer[rcount] & 0x1) == 1) ? true : false; // ??enabled flag?? 
			rcount++;
				
			int bptype = Trace32Line.getLongFromBuffer(rxBuffer, rcount);
			rcount+=4;
			
			int auxtype = Trace32Line.getLongFromBuffer(rxBuffer, rcount);
			rcount+=4;
				
			/* We do not report temporary and stepping breaks to Eclipse. */
			
			final int EMU_BREAKSOFT_TYPE_TEMP     =  0x00010000;
			final int EMU_BREAKSOFT_TYPE_STEPPING =  0x00020000;

			if ((auxtype & (EMU_BREAKSOFT_TYPE_TEMP | EMU_BREAKSOFT_TYPE_STEPPING)) !=0     ) {
				continue;
			}
						
			Trace32Breakpoint newBreak = new Trace32Breakpoint(this, byte_address, enabled, bptype);
			remoteBreaks.put(new Long(byte_address), newBreak);
				
			/* TODO: this code does not yet take into account different memory spaces D:, P: */
			{ 
				String pathAndFileName;
				int sourceLine;
							
				/* If break on same address exists, use its filename/linenumber, avoids RPC */
				if (knownBreaks.containsKey(new Long(byte_address))) {
					Trace32Breakpoint old = (Trace32Breakpoint)knownBreaks.get(new Long(byte_address));
					pathAndFileName = old.getFileName();
					sourceLine      = old.getLineNumber();
				} else {
					pathAndFileName = getSourceFile(newBreak.getAddress()); /* TODO Optimize: make single call to CAPI */
					sourceLine      = getSourceLine(newBreak.getAddress());
				}
				newBreak.setFileName(pathAndFileName);
				newBreak.setLineNumber(sourceLine);
			}
		}
		 	
		long trailer = Trace32Line.getLongFromBuffer(rxBuffer,rcount); /* ABCD; KILL */
		 	
		return remoteBreaks;
		 	
	}


	/**
	 * Executes the given PRACTICE command and checks if it succeeded. This is
	 * done by echoing the command to area window and reading back the message
	 * line. If the command echo is replaced by something else (an error
	 * message) it is interpreted as an error.
	 * 
	 * @throws Trace32Exception
	 *             if an error is detected; this is done by analysing the
	 *             message line.
	 * 
	 * @return contents of the message line i.e. last line in area window.
	 * 
	 * @note Only use this command for commands that do not produce output in
	 *       the area window (see above).
	 */

	public synchronized String checkedCommand(String pcommand)
			throws Trace32Exception {

		String echoCommand = "print \"" + echoString + pcommand + "\"";
		command(echoCommand); /* echo command to area window */

		command(pcommand); /* execute the command itself */

		/*
		 * Check last line of area window; if the echoString (<ECLIPSE>) was replaced
		 * by something else we assume an error.
		 */
		String answer = getMessage();
		if (answer.indexOf(echoString) == -1) {
			throw new Trace32Exception(Trace32Exception.PRACTICEERR);
		}

		return answer;
	}

	/**
	 * Executes a PRACTICE command in T32.
	 * 
	 * @param enableEcho
	 *            controls whether the command is printed to the area window
	 *            before execution.
	 * 
	 * @return value of the message line;
	 */
	
	public synchronized String uncheckedCommand(String pcommand,
			boolean enableEcho) throws Trace32Exception {

		if (enableEcho) {
			String echoCommand = "print \"" + echoString + pcommand + "\"";
			command(echoCommand); /* echo command to area window */
		}
		command(pcommand); /* execute the command itself */
		String answer = getMessage();
		return answer;
	}

	/** Convenience method to send modification notifications to all listeners. */
	protected void reportCreated(Trace32Breakpoint bp) {
		Enumeration listeners = breakpointListeners.elements();
		while (listeners.hasMoreElements()) {
			((ITrace32Listener) (listeners.nextElement()))
					.createdBreakpoint(bp);
		}
	}

	/** Convenience method to send modification notifications to all listeners. */
	protected void reportDeleted(Trace32Breakpoint bp) {
		Enumeration listeners = breakpointListeners.elements();
		while (listeners.hasMoreElements()) {
			((ITrace32Listener) (listeners.nextElement()))
					.deletedBreakpoint(bp);
		}
	}

	/** Convenience method to send modification notifications to all listeners. */
	protected void reportModified(Trace32Breakpoint bp) {
		Enumeration listeners = breakpointListeners.elements();
		while (listeners.hasMoreElements()) {
			((ITrace32Listener) (listeners.nextElement()))
					.modifiedBreakpoint(bp);
		}
	}

	/**
	 * Maps a file name to its module name.
	 * 
	 * @limitation Currently assumes 1:1 relation between module and file names
	 *             and does a very simplistic conversion:
	 *             pathAndFileName=="c:/etc/zsp500/sieve.c" -> module=="sieve"
	 * 
	 * @todo: either we make this function cleverer or adapt T32 to accept a
	 *        <filename:linenr>-syntax for setting breakpoints.
	 */
	private static String getModuleName(String pathAndFileName) {

		String fileName = new File(pathAndFileName).getName();
		int dotIndex = fileName.lastIndexOf('.');
		String moduleName = fileName.substring(0, dotIndex);

		return moduleName;
	}

	/** Dump all breaks onto console; for diagnostic purposes only.  */
	
	void dumpAllBreakpoints() throws Trace32Exception {
	
		Trace32Breakpoint[] a = getBreakpoints();
		System.out.println("\n== DUMP OF BREAKPOINTS ==");
		for (int i = 0; i < a.length; i++) {
			System.out.println(a[i]);
		}
		System.out.println("=======================\n");
	}

}

