
/* A breakpoint is always connected to a complete line.  There may be multiple 
  Trace32Breakpoint objects for the same source line (e.g. multiple ASM breaks), 
  if the breaks are set in the debugger and polled by the JAPI.

  - all breaks in T32 that hit the same HLL line go to the first address of the line, 
    there can only be a single HLL break on each line.
  - a break inside a HLL line will be shown at the address and mark the complete HLL line; 
    there can be multiple ASM-breaks at addresses inside a HLL line
  --> from perspective of Eclipse a HLL line can have a HLL and/or an ASM break.
*/

package com.lauterbach.trace32.api;

public class Trace32Breakpoint { 

	private Trace32API trace32API; /* keep track of associated debugger instance; for enable/disable */
	
	private int lineNumber;         /* lineNumber and fileName are only auxiliary fields; they */
	private String fileName;        /* are not considered part of the state of the breakpoint */
	
	private boolean enabled = true;  /* TODO: does not yet work for onchip databreaks! changes not detected */
	
	private long byte_address;      /* on some machines differs from word address */
	                                /* NOTE: we do not yet distinguish between memory ranges like D:, P:, ... */
	private int bptype;
	
	boolean HLLBreak=false, ASMBreak=false;  /* These are not mutually exclusive */
	
	public final static int EMU_BREAKSOFT_TYPE_P = 0x01;
	public final static int EMU_BREAKSOFT_TYPE_H = 0x02;
	public final static int EMU_BREAKSOFT_TYPE_S = 0x04;
	public final static int EMU_BREAKSOFT_TYPE_R = 0x08;
	public final static int EMU_BREAKSOFT_TYPE_W = 0x10;
	public final static int EMU_BREAKSOFT_TYPE_A = 0x20;
	public final static int EMU_BREAKSOFT_TYPE_B = 0x40;
	public final static int EMU_BREAKSOFT_TYPE_C = 0x80;
	public final static int EMU_BREAKSOFT_TYPE_D = 0x100;
	public final static int EMU_BREAKSOFT_TYPE_E = 0x200;
	public final static int EMU_BREAKSOFT_TYPE_ABCDE = 0x3e0;
	public final static int EMU_BREAKSOFT_TYPE_REGISTERREAD = 0x400;
	public final static int EMU_BREAKSOFT_TYPE_REGISTERWRITE = 0x800;
	public final static int EMU_BREAKSOFT_TYPE_MEMORYREAD =	0x1000;
	public final static int EMU_BREAKSOFT_TYPE_MEMORYWRITE = 0x2000;
	public final static int EMU_BREAKSOFT_TYPE_ALL = 0x3fff;
	
	/* Package access: breakpoints object are only created by the JAPI, either
	 * as response to API calls or because breakpoints are set in the debugger.
	 */

	Trace32Breakpoint(Trace32API _api, long _byte_addr, boolean _enabled, int _bptype) {
		trace32API   = _api;
		byte_address = _byte_addr;
		enabled      = _enabled;
		bptype       = _bptype;
	}

	/**
	 * Compares two breakpoints objects. Note that filename and linenumber are
	 * not compared. These are only auxiliary fields for the JAPI.
	 */
	public boolean equals(Object obj) {
		if (obj instanceof Trace32Breakpoint) {
			Trace32Breakpoint cmp = (Trace32Breakpoint) obj;
			return (byte_address == cmp.byte_address) 
				&& (enabled == cmp.enabled)
				&& (bptype  == cmp.bptype);	
		}
		return false;
	}

	public String toString() {
		return "Trace32Breakpoint <" + fileName +
			":" + Integer.toString(lineNumber) +
			": enabled=" + new Boolean(enabled) +
			": type=" + (isProgramBreak()? "PROGRAM" : "OTHER") +
			" >";
	}
	
	/* Only for internal use, currently */
	long getAddress() {
		return byte_address; 
	}
	
	/* Only for internal use, currently */
	void setAddress(long a){
		byte_address = a;
	}

	
	/* Returns the Trace32API to which the breakpoint is connected */
	public Trace32API getTrace32API() {
		return trace32API; 
	}
	
	
	/** Sets fileName. Has package access because only API can set this. */ 
	void setFileName(String pathAndFileName ) {
		fileName = pathAndFileName;		
	}
	/** Sets lineNumber. Has package access because only API can set this. */
	void setLineNumber(int sourceLine) {
		lineNumber = sourceLine;
	}
		
	/** Returns fully qualified file name */ 
	public String getFileName() {
		return fileName;
	}

	/** Returns line number on which the breakpoint is set; Line numbers start_line with 1 */
	public int getLineNumber() {
		return lineNumber;
	}
	
	/** Returns true for program breakpoints */
	public boolean isProgramBreak() {
		return (bptype &  EMU_BREAKSOFT_TYPE_P) != 0;
	}
	
	/** Returns true, if the breakpoint was set at the beginning of a HLL line.
	 * 
	 * Note: due to the merging of multiple breaks set inside a single source line, 
	 * both isHLLBreak() and isASMBreak() can be true simultaneously. This indicates 
	 * that a HLL break and at least one ASM break was set onto this line.  
	 */
	public boolean isHLLBreak() {
		return HLLBreak;
	}

	/** Returns true, if the breakpoint was set (possibly only in the debugger) onto a machine 
	 * instruction that is not at the beginning of a HLL line. 
	 */
	public boolean isASMBreak() {
		return ASMBreak;
	}

	/** Returns if breakpoint is enabled/disabled */
	public boolean isEnabled( ) {
		return enabled;
	}
	
	
	/* NOTE: Changes to the breakpoint state are _always_ done in class Trace32API and 
	 * reported to the JAPI after receiving status updates from T32.  
	 * This is done, because the behaviour of T32 is sometimes difficult to predict 
	 * (e.g. merging of breaks, ...) and we remain independent of T32 implementation changes.
	 * Actual state changes are _never_ done in class Trace32Breakpoint.
	 * 
	 * Trace32API may optimize in simple cases and report changes to listeners w/o polling 
	 * the new state from T32. If the optimization is wrong, the correct values from T32 
	 * will override it sooner or later. 
	 */
	
	/* Enables/disables breakpoint. The change will be reported to the breakpointListeners.
	 *  
	 * Note: T32 will enable/disable all breakpoints on the same source 
	 * line. Therefore, changing the state of a one breakpoint may trigger
	 * multiple changes notifications.
	 * 
	 * @param b selects enable/disable 
	 */
	public void enable(boolean b) throws Trace32Exception {
		trace32API.enableBreakpoint(this, b);  /* creates change notifications */
	}
	
	/* we migh move these into the constructor: there is hardly any need to modify the values later on */
	/* Accessor for use by Trace32API */
	void setEnabled(boolean p) {
		enabled = p;
	}
	/* Accessor for use by Trace32API */
	void setBPType(int p) {
		bptype = p;
	}
	
}
