/*
 * Created on 11.07.2003
 *
 */

package com.lauterbach.trace32.api;

/**
 * @author Rudi Dienstbeck
 *
 */

import java.io.*;
import java.net.*;

public class Trace32Line {
	
	public boolean up = false;            /* indicates if connected T32.exe */
	
	final static int PKG_LENMAX = 1472;
	final static int PKG_SYNCREQUEST = 2;
	final static int PKG_CONNECTREQUEST = 3;

	private DatagramSocket socket;        /* local socket */

	private InetAddress t32NodeAddr;      /* (remote) IP address of t32.exe process */
	// qui c'era localhost
	private String t32Node = "localhost"; /* Node on which the T32.exe process is executing */
	private int txPort     = 10000;       /* (remote) Port on which T32.exe listens for connections */
	private int rxPort     = 0;           /* (local) port to receive traffic */

	int packetSize = 1024;

	/* Attention: in T32 these variables are 16bit, but in Java 32bits.
	 * Take care to clip them off via &0xffff before comparing them */ 
	private int txSeqNum;   /* is incremented at each sent message */
	private int rxSeqNum;
	private boolean rxToggleBit;

	/* private int hostPort = 0;*/  	/* [!]DIE; what is this? seems unused */
	/* private int pollTime   = 5; */ 	/* [!]DIE; seems unused */


	private Trace32Line(){
		/* keep all default settings */
	}
	
	public Trace32Line(int _txPort) {
		txPort = 	_txPort;
	}
	
	public int getTxPort(){
		return txPort;
	};

	// --- configure --------------------------------------------------
	
	public void configure(String _node, int _txPort, int _packetSize) { 
		t32Node    = _node;    
		txPort	   = _txPort;
		packetSize = _packetSize;
	}

	/* @deprecated function. Use constructor instead */
	
	public void configure (String param, String value)
	{
		if (param.equals ("NODE="))
			t32Node = value;

		else if (param.equals ("PORT="))
			txPort = new Integer(value).intValue();
		
		else if (param.equals ("PACKLEN="))
			packetSize = new Integer(value).intValue();
	}
	
	// --- init --------------------------------------------------
	

	/* Connects to t32 process via socket interace.

	Silently returns, if already connected.

	@throws Trace32Exception
	
	*/
	public void init () throws Trace32Exception
	{
		if (up) return;
		
		try {
			t32NodeAddr = InetAddress.getByName(t32Node);

			socket = new DatagramSocket (); /* create local socket */
			
			//System.out.println("SoTimeout=" + socket.getSoTimeout());
			rxPort = socket.getLocalPort();
		
			if (socket.getReceiveBufferSize() > packetSize) /* reduce buffer sizes to required value. [!]DIE; should not we increase rather? */
				socket.setReceiveBufferSize(packetSize);
			if (socket.getSendBufferSize() > packetSize)
				socket.setSendBufferSize(packetSize);
			socket.setSoTimeout (500);
			
			socket.connect (t32NodeAddr, txPort);
			
			int i;
			for (i = 0; i < 10; i++) {
				if (connect()) {
					break;
				}
			}
			
			if (i == 10)
				throw new Trace32Exception (Trace32Exception.NORESPONSE);
		}
		catch (Trace32Exception e) {
			/* 		exit();  [!]  removed: we do not want to terminate application */
			throw e;
		}
		catch (UnknownHostException e)
		{
			throw new Trace32Exception (Trace32Exception.UNKNOWNHOST, t32Node);
		}
		catch (SocketException e)
		{
			throw new Trace32Exception (Trace32Exception.SOCKET, e.getMessage());
		}
		
		rxToggleBit = true;
		
		up = true;
	}

	// --- exit --------------------------------------------------
	
	public void exit ()
	{
		byte[] buffer = new byte[PKG_LENMAX];
		final byte[] magic = {'T','R','A','C','E','3','2',0};
		DatagramPacket packet = 
			new DatagramPacket (buffer, packetSize, t32NodeAddr, txPort);
		
		buffer[0] = 4;
		buffer[1] = 0;

		System.arraycopy (magic, 0, buffer, 8, 8);

		if (socket != null)
		{
			try {
				for (int i = 0; i < 5; i++)
					socket.send (packet);
			}
			catch (IOException e) {
				;	// let API exit, even at failure
			}

			socket.disconnect();
			socket.disconnect();
			socket.close();
			socket.close();
		}
		
		socket = null;
		up = false;
	}
	
	// --- transmit --------------------------------------------------
	
	public void transmit (byte[] buf, int len) throws Trace32Exception
	{
		// LINE_Transmit
		byte[] buffer = new byte[len+5];
		buffer[0] = 0;
		buffer[1] = 0;
		buffer[2] = 0;
		buffer[3] = 0;
		buffer[4] = 0;
		System.arraycopy (buf, 0, buffer, 5, len);
		txIntern (buffer, len+5);
	}

	private void txIntern (byte[] buf, int size) throws Trace32Exception
	{
		// LINE_LineTransmit
		byte[] buffer = new byte[packetSize];
		int srcIndex = 0;
		
		do {
			int txsize;
			if (size > packetSize - 4)
				txsize = packetSize - 4;
			else
				txsize = size;
				
			buffer[0] = 0x11;	// RESPONSE
			buffer[1] = (byte) ((size > txsize) ? 1 : 0);	/* packets follow */
			setBufferFromWord (buffer, 2, txSeqNum);
			System.arraycopy (buf, srcIndex, buffer, 4, txsize);
			try {
				socket.send (new DatagramPacket (buffer, txsize+4, t32NodeAddr, txPort));
			}
			catch (IOException e) {
				throw new Trace32Exception (Trace32Exception.IO, e.getMessage());
			}
			
			txSeqNum++;
			srcIndex += txsize;
			size -= txsize;
			
		} while (size > 0);
	}
	
	// --- receive --------------------------------------------------
	
	public byte[] receive (int msgSeqNum) throws Trace32Exception
	{
		byte[] buffer = new byte[packetSize];
		final byte MSG_LRETRY  = 0x08;
		final byte MSG_LHANDLE = 0x10;
		
		for (int retry = 0; retry < 5; retry++) 
		{
			buffer = rxIntern ();
			
			if (buffer == null)		// timeout
				continue;
				
			if (buffer[3] == 0xfe) {
				retry = 0;
				rxToggleBit = true;
				continue;
			}
			if (buffer[4] != msgSeqNum)
				continue;
			if ((buffer[0] & MSG_LRETRY) != 0)
				if (rxToggleBit == ((buffer[0] & MSG_LHANDLE) != 0))
				{
					try {
						byte[] buf = new byte[0];
						transmit (buf, 0);
					}
					catch (Trace32Exception e) {
						if (e.getCause() == Trace32Exception.IO)
							sync();
						else throw e;
					}
					continue;
				}
			rxToggleBit = ((buffer[0] & MSG_LHANDLE) != 0);
			
			return buffer;
		}
		
		throw new Trace32Exception (Trace32Exception.RECEIVE);
	}
	
	private byte[] rxIntern () throws Trace32Exception
	{
		byte[] buffer = new byte[PKG_LENMAX];
		byte[] buf1 = new byte[0];
		byte[] buf2;
		byte[] resultBuffer;
		byte flag;
		int count = 0;
		final byte[] handshake = {7,0,0,0,0,0,0,0,'T','R','A','C','E','3','2',0};
		DatagramPacket packet = 
			new DatagramPacket (buffer, packetSize);
			
		do {
			do {
				try {
					socket.receive (packet);
				}
				catch (InterruptedIOException e)	// timeout
				{
					return null;
				}
				catch (IOException e) {
					throw new Trace32Exception (Trace32Exception.IO, e.getMessage());
				}
				if (buffer[0] != 0x01)
					throw new Trace32Exception (Trace32Exception.RECEIVE, "(intern)");
					
			} while (getWordFromBuffer(buffer, 2) != ( rxSeqNum & 0xffff));
			
			rxSeqNum = (rxSeqNum+1)&0xffff;
			flag = buffer[1];
			buf2 = buf1;
			buf1 = new byte [count + packet.getLength()];
			System.arraycopy(buf2, 0, buf1, 0, count);
			System.arraycopy(buffer, 4, buf1, count, packet.getLength()-4);
			count += packet.getLength() - 4;

			if (count > (16384-256)) 
				throw new Trace32Exception (Trace32Exception.RECEIVE, "(length)");
				
			if (flag == 2) 
				try {
					socket.send (new DatagramPacket (handshake, 16, t32NodeAddr, txPort));
				}
				catch (IOException e)
				{
					throw new Trace32Exception (Trace32Exception.IO, e.getMessage());
				}
				
		} while (flag != 0);
		
		resultBuffer = new byte[count];
		System.arraycopy (buf1, 0, resultBuffer, 0, count);
		
		return resultBuffer;
	}
	
	// --- sync --------------------------------------------------
	
	public void sync() throws Trace32Exception
	{
		int retry;
		
		for (retry = 0; retry < 5; retry++)
			if (syncIntern() == true)
				return;
			
		throw new Trace32Exception (Trace32Exception.SYNC);
	}
	
	private boolean syncIntern() throws Trace32Exception
	{
		byte[] rxBuffer = new byte[PKG_LENMAX];
		byte[] txBuffer = new byte[PKG_LENMAX];
		final byte[] magic = {'T','R','A','C','E','3','2',0};
		byte[] compare = new byte[8];
		DatagramPacket rxPacket = 
			new DatagramPacket (rxBuffer, packetSize);
		DatagramPacket txPacket = 
			new DatagramPacket (txBuffer, packetSize, t32NodeAddr, txPort);
		boolean retry;

		try {
			do {
				int j;
				retry = false;
					
				txBuffer[0] = PKG_SYNCREQUEST;
				txBuffer[1] = 0;
				setBufferFromWord (txBuffer, 2, txSeqNum);
				setBufferFromWord (txBuffer, 4, 0);
				setBufferFromWord (txBuffer, 6, 0);
				System.arraycopy (magic, 0, txBuffer, 8, 8);
				txPacket.setLength(16);
				socket.send (txPacket);

				for (j = 0; j < 20; j++) {		/* empty queue */
			
					socket.receive (rxPacket);
			
					System.arraycopy (rxBuffer, 8, compare, 0, 8);
			
					if ((rxPacket.getLength() != 16) || (rxBuffer[0] != 0x12) || 
						!(new String(compare).equals(new String(magic))))
					{
						if ((rxPacket.getLength() == 16) && (rxBuffer[0] == 5))
						{
							retry = true;
							break;
						}
						continue;
					}
			
					break;
				}
			
				if (j == 20)
					return false;	// failed
			} while (retry);

			rxSeqNum = getWordFromBuffer (rxBuffer, 2);

			txBuffer[0] = 0x22;		/* SYNCBACK */
			txBuffer[1] = 0;

			setBufferFromWord (txBuffer, 2, txSeqNum);
			setBufferFromWord (txBuffer, 4, 0);
			setBufferFromWord (txBuffer, 6, 0);
			System.arraycopy (magic, 0, txBuffer, 8, 8);
			txPacket.setLength(16);

			socket.send (txPacket);
		}
		catch (IOException e) {
			throw new Trace32Exception (Trace32Exception.IO, e.getMessage());
		}
		
		return true;	// passed
	}
	
	// --- connect --------------------------------------------------
	
	private boolean connect () throws Trace32Exception
	{
		byte[] rxBuffer = new byte[PKG_LENMAX];
		byte[] txBuffer = new byte[PKG_LENMAX];
		//final byte[] magic = new String("TRACE32").getBytes("ISO-8859-1");
		final String magic = "TRACE32";
		DatagramPacket rxPacket = 
			new DatagramPacket (rxBuffer, packetSize);
		DatagramPacket txPacket = 
			new DatagramPacket (txBuffer, packetSize, t32NodeAddr, txPort);
		
		txBuffer[0] = PKG_CONNECTREQUEST;
		txBuffer[1] = 0;
		
		txSeqNum = 1;
		setBufferFromWord (txBuffer, 2, txSeqNum);
		setBufferFromWord (txBuffer, 4, txPort);
		setBufferFromWord (txBuffer, 6, rxPort);
		try {
			System.arraycopy (magic.getBytes("ISO-8859-1"), 0, txBuffer, 8, 7);
		}
		catch (Exception e)
		{
			throw new Trace32Exception (Trace32Exception.OTHER, e.getMessage());
		}
		txBuffer[15] = 0;
		
		try {
			socket.send (txPacket);
			socket.receive (rxPacket);
		}
		catch (InterruptedIOException e)	// timeout
		{
			return false;
		}
		catch (IOException e) {
			throw new Trace32Exception (Trace32Exception.IO, e.getMessage());
		}
		String compare = new String (rxBuffer, 8, 7);
		if (!compare.equals(magic))
			return false;
			
		rxSeqNum = getWordFromBuffer (rxBuffer, 2);
		
		if (rxBuffer[0] == 0x53)
			throw new Trace32Exception (Trace32Exception.ACCESS);
			
		if (rxBuffer[0] != 0x13)
		{
			if (rxBuffer[0] == 0x23)
				throw new Trace32Exception (Trace32Exception.ACCESS);
			return false;
		}
		
		packetSize = rxPacket.getLength();
		
		return true;
	}
	
	/* Helper methods. Move to some util class or replace by Java API calls */
	
	/* for array<->int conversions see 
	http://forum.java.sun.com/thread.jspa?threadID=628082&messageID=3598953
	*/
		
	static int getWordFromBuffer (byte[] buffer, int index)
	{
		return unsignedByteToInt(buffer[index]) + (unsignedByteToInt(buffer[index+1]) << 8);
	}

	/* NOTE: we have a problem with signedness here: 
	 * java types are always SIGNED, so 32bit unsigned will not fit into a java-int. We need to adapt the data type somewhat */
	static int getLongFromBuffer (byte[] buffer, int index)
	{  
		return unsignedByteToInt(buffer[index]) 
		    + (unsignedByteToInt(buffer[index+1]) << 8)
			+ (unsignedByteToInt(buffer[index+2]) << 16)
			+ (unsignedByteToInt(buffer[index+3]) << 24);
	}
	
	static void setBufferFromWord (byte[] buffer, int index, int value)
	{
		buffer[index]   = (byte) (value & 0xff);
		buffer[index+1] = (byte) ((value >> 8) & 0xff);
	}
	static void setBufferFromLong (byte[] buffer, int index, int value)
	{
		buffer[index]   = (byte) (value & 0xff);
		buffer[index+1] = (byte) ((value >> 8)  & 0xff);
		buffer[index+2] = (byte) ((value >> 16) & 0xff);
		buffer[index+3] = (byte) ((value >> 24) & 0xff);
	}
	
	
	static int unsignedByteToInt (byte b) {
		return (int) b & 0xFF;
	}
}
