/** 
 * commentata
 */
package cpu.nios2;

import cpu.CPU;

/** Rappresenta una CPU NiosII
 * @author Pierfrancesco
 *
 */
public class Nios2CPU extends CPU{
	
	// codici dei branch per la CPU NiosII
	public static final int BNE=0x1E;
	public static final int BLT=0x16;
	public static final int BEQ=0x26;
	public static final int BGE=0x0E;
	public static final int BGEU=0x2E;
	public static final int BLTU=0x36;
	public static final int BR=0x06;		// salto incondizionato
	public static final int RET=0xF800283A;
	public static final int CALLR=0x003EE83A;
	public static final int CALL=0x00000000;
	
	/** fattore di conversione da cicli di clock a microsecondi
	 *  per NiosII
	 */
	public static final double TO_US=(0.000078125D);
	
	/** fattore di conversione da cicli di clock a millisecondi
	 *  per NiosII
	 */
	public static final double TO_MS=(0.000000078125D);
	
	
	/** Maschere per trovare gli opcode e le destinazioni
	 * 
	 */
	public static final int MASK_OPCODE_JUMP=0x0000003F;
	public static final int MASK_DEST_JUMP=0x003FFFC0;
	
	
	/** Restituisce la lunghezza (in byte) delle istruzioni del NiosII 
	 * @return			4	bytes
	 */
	public static int lengthInstr() {
		return 4;	// restituisce 4 bytes perche' il RISC NiosII
					// ha tutte le istruzioni di 32 bit
	}
	
	/** Converte da numero di cicli in microsecondi
	 * @param time			num di cicli di clock
	 * @return
	 */
	public static double to_us(long time) {return (double) ((double)time * TO_US);}
	
	/** Converte da numero di cicli in nanosecondi
	 * @param time			num di cicli di clock
	 * @return
	 */
	public static double to_ms(long time) {return (double) ((double) time * TO_MS);}
	
	/** 
	 * @param bus_data_width 
	 * @param bus_data_long 
	 * @return		restituisce l'istruzione del corrente record
	 */
	public static int getCurrInstr(int bus_data_width, long bus_data_long) {
		
		final long MASK1=0x000000FFL;
		final long MASK2=0x0000FFFFL;
		final long MASK3=0x00FFFFFFL;
		final long MASK4=0x00000000FFFFFFFFL;
		
		long mask;
		// a seconda della lunghezza dei bytes fetchati
		switch (bus_data_width) {
		case 1: mask=MASK1;
		break;
		case 2: mask=MASK2;
		break;
		case 3: mask=MASK3;
		break;
		case 4: mask=MASK4;
		break;
		default:mask=MASK4;
		break;
		}
		// maschero per togliere via i byte non significativi
		return (int) (bus_data_long & mask);
		
	}
	
	
	/**
	 * @return		restituisce l'opcode del jump (se e' un jump)
	 */
	public static int getOpCodeJump(int bus_data_width, long bus_data_long) {
		long curr_instr=getCurrInstr(bus_data_width, bus_data_long);
		// maschero per avere l'opcode
		int curr_opcode=(int) (curr_instr & Nios2CPU.MASK_OPCODE_JUMP);
		return curr_opcode;
	}
	
	/** Controlla se il record descrive un salto (qualunque)
	 * @return				true se rappresenta un salto (qualunque)
	 */
	public static boolean isJump(long opcode) {
		
		if (opcode==Nios2CPU.BR)
			return true;
		else
			if (isCondJump(opcode))
				return true;
			else
				return false;
	}
	
	public static boolean isRet( long instr) {
		long opcode= instr;
		if (opcode==RET)
			return true;
		else
			return false;
	}
	
	public static boolean isCall(long instr) {
		long opcode1=instr & 0x000000000000003FL;
		long opcode2=instr & 0x0000000007FFFFFFL;
		if ( (opcode1==CALL) | (opcode2==CALLR) )
			return true;
		else
			return false;
		
		
	}
	
	/** Controlla se il record descrive un salto condizionato
	 * @return				true se rappresenta un salto condizionato
	 */
	
	public static boolean isCondJump(long opcode) {
		if ((opcode==Nios2CPU.BNE) |
			(opcode==Nios2CPU.BLT) |
			(opcode==Nios2CPU.BEQ) |
			(opcode==Nios2CPU.BGE) |
			(opcode==Nios2CPU.BGEU)|
			(opcode==Nios2CPU.BLTU) )
			return true;
		else
			return false;
		
		/*switch (opcode) {
		
		case Nios2CPU.BNE:
		case Nios2CPU.BLT:
		case Nios2CPU.BEQ:
		case Nios2CPU.BGE:
		case Nios2CPU.BGEU:
		case Nios2CPU.BLTU:	
			return true;
		default: return false;	
		}*/
		
		
	}
	
	/** restituisce l'indirizzo di destinazione del salto
	 * @return				indirizzo di destinazione del salto
	 */
	public static long getDestJump(int bus_data_width, long bus_data_long, long paddress) {
		int imm16;
		long curr_instr= getCurrInstr(bus_data_width, bus_data_long);
		// maschero i bit della destinazione
		imm16=(int) ((curr_instr & Nios2CPU.MASK_DEST_JUMP) >> 6);
		// li shifto a destra di 6 bit per averli allineati

		int curr_offset;
		if ((imm16 & 0x00008000)!=0)
			curr_offset=(imm16| 0xFFFF0000);
		else
			curr_offset=imm16;
		
		return  (int) (curr_offset+paddress+Nios2CPU.lengthInstr());	
		
	}
	
	public static long getDestCall(int bus_data_width, long bus_data_long ) {
		
		
	int imm26;
	long curr_instr= getCurrInstr(bus_data_width, bus_data_long);
	imm26= (int) (curr_instr & 0xFFFFFFC0);
	long addr_called=(imm26 >> 4);
	return addr_called;
	
	}
	
	public static long allinea(long addr) {
		long resto = addr % Nios2CPU.lengthInstr();
		return (addr-resto);
	
	}
	
}
